/*
 * mapping.c - routines for recording, displaying walked path.
 */

#include <stdio.h>
#include <string.h>
#include <unistd.h>

#include "cancan.h"

static char reverse_dir();

/*
 * mapping variables
 */
static char mappath[MAXMAPLEN]; /* circular path list */
static int mapstart = 0;	/* index to first map entry */
static int mapend = 0;		/* index one past last map entry */

#define MAPINDEX(i) (((i) + MAXMAPLEN) % MAXMAPLEN)
#define MAPENTRY(i) mappath[MAPINDEX(i)]

/*
 * return reverse direction
 */
static char reverse_dir(dir)
char dir;
{
    static char dirs[] = "nsewud";
    char *p = strchr(dirs, dir);
    return p ? dirs[(p - dirs) ^ 1] : 0;
}

/*
 * remove steps from map end or beginning.
 */
void remove_steps(steps)
int steps;
{
    int mapsize = MAPINDEX(mapend - mapstart);

    if(steps < 0) {
	mapend = mapsize + steps > 0 ? MAPINDEX(mapend + steps) : mapstart;
    } else if(mapsize > steps) {
	mapstart = MAPINDEX(mapend - steps);
    }
}

/*
 * retrace steps on map, and walk them back
 */
void retrace_steps(steps)
int steps;
{
    static char cmd[] = "x\n";
    int mapsize = MAPINDEX(mapend - mapstart);

    if(steps <= 0) {
	if((steps += mapsize) < 0) return;
    } else if(steps > mapsize) {
	steps = mapsize;
    }
    while(steps--) {
	mapend = MAPINDEX(mapend - 1);
	cmd[0] = reverse_dir(mappath[mapend]);
	send_to_main_host(cmd, 2);
    }
}

/*
 * show automatic map (latest steps) of the form s2ews14n
 */
void show_automap()
{
    char lastdir;
    int count = 0;
    int i;

    if(mapstart == mapend) {
	printf("#map empty\n");
    } else {
	printf("#map: ");
	lastdir = mappath[mapstart];
	for(i = mapstart; i != mapend; i = MAPINDEX(i + 1)) {
	    if(mappath[i] != lastdir) {
		if(count > 1)
		    printf("%d", count);
		putchar(lastdir);
		count = 1;
		lastdir = mappath[i];
	    } else
		count++;
	}
	if(count > 1)
	    printf("%d", count);
	printf("%c\n", lastdir);
    } 
}

/*
 * add direction to automap
 */
void add_dir_to_map(dir)
char dir;
{
    if(mapend != mapstart && dir == reverse_dir(MAPENTRY(mapend - 1))) {
	mapend = MAPINDEX(mapend - 1); /* retrace one step */
    } else {
	MAPENTRY(mapend) = dir;
	mapend = MAPINDEX(mapend + 1);
	if(mapend == mapstart)
	    mapstart = MAPINDEX(mapstart + 1);
    }
}
